import struct


class DirEntry:
    DTYPE_STR = ('DEL', 'SEQ', 'PRG', 'USR', 'REL', '???', '???', '???')

    def __init__(self, image, name, file_type, disk_type, start_addr, end_addr, img_start):
        self.image = image
        self.name = name
        self.file_type = file_type
        self._disk_type = disk_type
        self.start_addr = start_addr
        self.end_addr = end_addr
        self.img_start = img_start

    @property
    def disk_type(self):
        return self.DTYPE_STR[self._disk_type & 7]

    @property
    def protected(self):
        return False

    @property
    def closed(self):
        return True

    def contents(self):
        data = b''
        if self.disk_type == 'PRG':
            data = struct.pack('<H', self.start_addr)
        img_end = self.img_start+self.end_addr-self.start_addr
        data += self.image.map[self.img_start:img_end]
        return data



import unittest

from t64.dir_entry import DirEntry


class TestDirEntry(unittest.TestCase):

    def setUp(self):
        self.entry = DirEntry(None, b'TEST', 1, 0x82, 0x1000, 0x2000, 0x3000)

    def test_file_type(self):
        self.assertEqual(self.entry.file_type, 1)

    def test_disk_type(self):
        self.assertEqual(self.entry.disk_type, 'PRG')

    def test_protected(self):
        self.assertFalse(self.entry.protected)
        self.entry._disk_type = 0xc2
        self.assertFalse(self.entry.protected)

    def test_closed(self):
        self.assertTrue(self.entry.closed)
        self.entry._disk_type = 0x02
        self.assertTrue(self.entry.closed)



import codecs
import logging
import math
import mmap
import struct

from pathlib import Path

from .dir_entry import DirEntry
from .magic import is_valid_image


log = logging.getLogger(__name__)


class T64ImageReader:
    def __init__(self, filepath):
        self.filepath = Path(filepath) if isinstance(filepath, str) else filepath
        self.fileh = None
        self.map = None

    def open(self):
        if not is_valid_image(self.filepath):
            raise ValueError("Invalid file format")
        self.fileh = self.filepath.open('rb')
        self.map = mmap.mmap(self.fileh.fileno(), 0, mmap.MAP_PRIVATE, mmap.ACCESS_READ)

        self.version, self.max_entries, self.used_entries = struct.unpack('<HHH', self.map[0x20:0x26])
        self.tape_name = self.map[0x28:0x40]
        if self.max_entries == 0:
            raise ValueError("Invalid max entries, 0")
        if self.used_entries == 0:
            raise ValueError("Invalid used entries, 0")

        self.entries = []
        for eidx in range(0, self.max_entries):
            estart = 0x40 + (eidx * 0x20)
            ftype, disk_type, start, end, _, img_start = struct.unpack('<BBHHHI', self.map[estart:estart+12])
            if ftype == 1:
                if end == 0xc3c6 or end <= start:
                    raise ValueError("Invalid entry end address, 0x{:x}".format(end))
                file_name = self.map[estart+0x10:estart+0x20].rstrip(b' ')
                dir_entry = DirEntry(self, file_name, ftype, disk_type, start, end, img_start)
                self.entries.append(dir_entry)
            elif ftype != 0:
                log.warning("Unsupported file type, "+str(ftype))

    def close(self):
        if self.map:
            self.map.close()
        if self.fileh:
            self.fileh.close()

    def directory(self, pattern=b'', encoding='petscii-c64en-uc', drive=0):
        try:
            _ = codecs.lookup(encoding)
        except LookupError:
            log.warning("PETSCII codecs not available, using ASCII")
            encoding = 'ascii'

        yield '{} "{:24}"'.format(drive, self.tape_name.decode(encoding))

        for entry in self.entries:
            if entry.name.startswith(pattern):
                size_blocks = math.ceil((entry.end_addr-entry.start_addr) / 254)
                closed_ch = ' ' if entry.closed else '*'
                disk_type = entry.disk_type
                if entry.protected:
                    disk_type += '<'
                name = '"'+entry.name.decode(encoding)+'"'
                yield '{!s:5}{:18}{}{}'.format(size_blocks, name, closed_ch, disk_type)

    def iterdir(self):
        yield from self.entries

    def entry(self, name):
        entry_list = [e for e in self.entries if e.name == name]
        if entry_list:
            return entry_list[0]
        return None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __str__(self):
        return "<{}: {!s}>".format(type(self).__name__, self.filepath)



import sys

from t64 import T64ImageReader


with T64ImageReader(sys.argv[1]) as t:
    print([e.name for e in t.entries])
    with open(sys.argv[3], 'wb') as fileh:
        e = t.entry(sys.argv[2].encode())
        fileh.write(e.contents())



import unittest
from pathlib import Path
from unittest.mock import patch, mock_open, Mock

from t64.t64_image_reader import T64ImageReader


@patch('t64.t64_image_reader.is_valid_image')
class TestT64ImageReader(unittest.TestCase):

    def setUp(self):
        self.map = bytearray(b'C64 tape image file'.ljust(32, b'\x00'))
        self.map += b'\x01\x01\x02\x00\x01\x00\x00\x00'
        self.map += b'tape name'.ljust(24, b' ')
        self.map += b'\x00'*32
        with patch.object(Path, 'open', mock_open(read_data=bytes(self.map))):
            self.image = T64ImageReader(Path())
        self.image.filepath = Mock()

    def test_header(self, mock_valid):
        mock_valid.return_value = True
        self.map += b'\x00'*32
        with patch('t64.t64_image_reader.mmap.mmap') as mock_mmap:
            mock_mmap.return_value = self.map
            self.image.open()
        self.assertEqual(self.image.tape_name, b'tape name'.ljust(24, b' '))
        self.assertEqual(self.image.max_entries, 2)
        self.assertEqual(self.image.used_entries, 1)

    def test_entries(self, mock_valid):
        mock_valid.return_value = True
        self.map += b'\x01\x82\x00\x10\x00\x20'.ljust(16, b'\x00')
        self.map += b'file name'.ljust(16, b' ')
        with patch('t64.t64_image_reader.mmap.mmap') as mock_mmap:
            mock_mmap.return_value = self.map
            self.image.open()
        self.assertEqual(len(self.image.entries), 1)
        self.assertIsNotNone(self.image.entry(b'file name'))
        self.assertIsNone(self.image.entry(b'not found'))

    def test_directory(self, mock_valid):
        mock_valid.return_value = True
        self.map += b'\x01\x82\x00\x10\x00\x20'.ljust(16, b'\x00')
        self.map += b'file name'.ljust(16, b' ')
        with patch('t64.t64_image_reader.mmap.mmap') as mock_mmap:
            mock_mmap.return_value = self.map
            self.image.open()
        dir_list = [l for l in self.image.directory(encoding='ascii')]
        self.assertEqual(dir_list, ['0 "tape name               "', '17   "file name"        PRG'])




T64_MAGICS = (
    b'C64 tape image file',
    b'C64S tape file',
    b'C64S tape image file'
)


def is_valid_image(file_path):
    with file_path.open('rb') as fileh:
        magic = fileh.read(32)
    magic = magic.rstrip(b'\x00')
    return any([magic.startswith(m) for m in T64_MAGICS])



import unittest

from pathlib import Path
from unittest.mock import patch, mock_open

from t64.magic import is_valid_image


class TestMagic(unittest.TestCase):

    def test_good_magic(self):
        with patch.object(Path, 'open', mock_open(read_data=b'C64S tape image file'.ljust(32, b'\x00'))):
            self.assertTrue(is_valid_image(Path()))

    def test_bad_magic(self):
        with patch.object(Path, 'open', mock_open(read_data=b'INVALID')):
            self.assertFalse(is_valid_image(Path()))
